// Original Author: Matt Kropp, June 2000
//
// All files and code in this projected are
// subject to the GNU General Public License.
// A copy of this license may be found here:
// http://www.gnu.org/licenses/gpl.html
//
// Notes:
// Although Telnet looks simple from a 
// user's point of view, it is an extremely
// complex protocol. This code does not
// attempt to be a complete telnet client.
// Instead, it attempts to to the bare minimum
// required to control a telnet session from
// within an application.


#include <windows.h>
#include <userint.h>
#include <ansi_c.h>
#include <utility.h>
#include "telnet.h"


// Parses telnet commands
// This function is not necessary to define if it's not needed.
// Alter this function for any special option handling
void parseCommand(TelnetSession session, int doResponse, unsigned char *command)
{
	int lookForSE = 0;
	int gotSend = 0;
	const char *desc;
	int action = WONT;
	int option;
	unsigned char *ptr = command;

	
	if(*ptr != IAC) return;
	printf("IAC ");
	
	switch(*(++ptr))
	{
		// For this example all options are refused
		case DO:
			printf("DO ");
			action = WONT;
			break;

		case DONT:
			printf("DONT ");
			action = WONT;
			break;

		case WILL:
			printf("WILL ");
			action = DONT;
			break;

		case WONT:
			printf("WONT ");
			action = DONT;
			break;
				
		// Print out these commands for the user to see
		case SB:
			printf("SB ");
			// Set flag for subnegotioation command start
			lookForSE = 1;
			break;
			
		case DATA_MARK:
			printf("DATA_MARK\n");
			return;
			
		case EOR:
			printf("EOR\n");
			return;

		case BREAK:
			printf("BREAK\n");
			return;

		case INTERRUPT:
			printf("INTERRUPT\n");
			return;
		
		case ABORT_OUTPUT:
			printf("ABORT_OUTPUT\n");
			return;

		case ARE_YOU_THERE:
			printf("ARE_YOU_THERE\n");
			return;

		case ERASE_CHAR:
			printf("ERASE_CHAR\n");
			return;
		
		
		case ERASE_LINE:
			printf("ERASE_LINE\n");
			return;
		
		case GO_AHEAD:
			printf("GO_AHEAD\n");
			return;
			
		// Print the command number for all other commands
		default:
			printf("%u\n", *ptr);
			return;
	}

	option = 255;
	
	// Scan rest of buffer for closing command
	do
	{
		switch(*(++ptr))
		{
			case IAC:
				printf("IAC ");
				break;

			case SE:
				printf("SE");
				// Clear flag now that end of subnegotioation was found
				lookForSE = 0;
				break;
				
			case 1:
				if(lookForSE && !gotSend)
				{
					printf("SEND ");
					gotSend = 1;
				}
				else
				{
					printf("Echo\n");
					option = OPT_ECHO;
					if(action == DONT) action = DO;
				}
				break;
			
			case 0:
				if(gotSend)
				{
					printf("IS ");
					gotSend = 0;
				}
				else
				{
					printf("Binary Transmission\n");
					option = OPT_BINARY;
				}
					
				break;
				
				
			default:
				option = *ptr;
				telnet_GetOptionInfo (option, &desc, NULL);
					
				if(desc == NULL)
					printf("%u\n", *ptr);
				else
					printf("%s\n", desc);
					
				break;
		}

	}while(lookForSE);
	
	printf("\n");

	// Send a response if needed
	if(doResponse)
	{
			telnet_SendCommand(session, action, option);
	}
}

// Telnet session event handler
void myCallback(TelnetSession sessionHandle, 
				telnet_event TelnetEvent, 
				const unsigned char EventData[])
{

	switch(TelnetEvent)
	{
		case TELNET_DISCONNECT:
			printf("\nDisconnected\n");
			break;
			
		case TELNET_CONNECT: 
			printf("\nConnected\n");
			break;

		case TELNET_RECEIVE:
			printf((char *) EventData);
			break;
			
		case TELNET_SEND:
			break;
			
		case TELNET_CMD_RECEIVE:
			printf("\nCommand received\n");
			parseCommand(sessionHandle, 1, (unsigned char *)EventData);
			break;
		
		case TELNET_CMD_RESPONSE:
			printf("\nResponse Sent\n");
			parseCommand(sessionHandle, 0, (unsigned char *)EventData);
			break;
		
		default:
			break;
	}
}
                    
// Example of opening a telnet session and handling I/O
int main (int argc, char *argv[])
{
	int status;
	char *response = NULL;
	TelnetSession session, test;
	
	if (InitCVIRTE (0, argv, 0) == 0)
		return -1;    /* out of memory */
	
	// Try to open a telnet session to the localhost
	status = telnet_OpenSession (&session, "localhost", myCallback);
	printf("\nOpenSession: %s\n", telnet_GetErrorText (status));
		
	ProcessSystemEvents();
	
	// Send the don't suppress go-ahead option
	status = telnet_SendCommand (session, DONT, OPT_SUPPRESS_GO_AHEAD);
	printf("\nDON'T SUPPRESS GO-AHEAD: %s\n", telnet_GetErrorText (status));
	
	ProcessSystemEvents();
	
	// Handle login into server
	status = telnet_LogIn (session, "ogin:", "assword:", "user\r", "password\r");
	printf("\nLogged In: %s\n", telnet_GetErrorText (status));
	
	ProcessSystemEvents();
	
	// Send a ls command
	status = telnet_SendText (session, "dir\r");
	printf("\nSendText: %s\n", telnet_GetErrorText (status));

	ProcessSystemEvents();
	
	// Wait 5 seconds for response
	status = telnet_WaitForResponse (session, &response, 5.0);
	printf("\nGetResponse: %s\n", telnet_GetErrorText (status));

	// Show the response
	if(response != NULL)
	{
		printf("\nCommand Output:\n%s\n", response);
		free(response);
	}
	else
		printf("\nNo reponse received.\n");
	
	ProcessSystemEvents();
	
	// Close the telnet session
	status = telnet_CloseSession (&session);
	printf("\nCloseSession: %s\n", telnet_GetErrorText (status));
	

	return 0;
}
