/*======================================================================
* Original Author: Matt Kropp, June 2000
*
* All files and code in this projected are
* subject to the GNU General Public License.
* A copy of this license may be found here:
* http://www.gnu.org/licenses/gpl.html
*
* Notes:
* Although Telnet looks simple from a 
* user's point of view, it is an extremely
* complex protocol. This code does not
* attempt to be a complete telnet client.
* Instead, it attempts to to the bare minimum
* required to control a telnet session from
* within an application.
*
* This file may be compiled seperately as a DLL
*
========================================================================
Module:	Telnet Function Panel Code

Description:
		Contains functions for creating telnet sessions under 
		programmatic control.
========================================================================*/

#include <windows.h>
#include <ansi_c.h>
#include <formatio.h>
#include <utility.h>
#include <userint.h>
#include <tcpsupp.h>

#include "telnet.h"


// Default definitons
#define TELNET_PORT			  23
#define START_BUFFER_SIZE	8096	/* 8KB buffer */
#define MIN_BUFFER_SIZE		1460	/* Minimum remaining buffer size */
#define CONNECT_TIMEOUT		5000	/* 5 seconds */
#define PROMPT_TIMEOUT		5000	/* 5 seconds */
#define RECEIVE_TIMEOUT		3000	/* 3 seconds */
#define MORE_DATA_TIMEOUT	 200	/* 200 mS    */


// Telnet Options
typedef struct {
	unsigned char	option;
	const char *	description;
	const char *	RFC;
} telnet_option;

// Telnet Option info
telnet_option TelnetOptions[] = {

{   0,     "Binary Transmission",                                "RFC856" },
{   1,     "Echo",                                               "RFC857" },
{   2,     "Reconnection",                                 "NIC50005,JBP" },
{   3,     "Suppress Go Ahead",                              "RFC858,JBP" },
{   4,     "Approx Message Size Negotiation",              "ETHERNET,JBP" },
{   5,     "Status",                                         "RFC859,JBP" },
{   6,     "Timing Mark",                                    "RFC860,JBP" },
{   7,     "Remote Controlled Trans and Echo",               "RFC726,JBP" },
{   8,     "Output Line Width",                            "NIC50005,JBP" },
{   9,     "Output Page Size",                             "NIC50005,JBP" },
{  10,     "Output Carriage-Return Disposition",             "RFC652,JBP" },
{  11,     "Output Horizontal Tab Stops",                    "RFC653,JBP" },
{  12,     "Output Horizontal Tab Disposition",              "RFC654,JBP" },
{  13,     "Output Formfeed Disposition",                    "RFC655,JBP" },
{  14,     "Output Vertical Tabstops",                       "RFC656,JBP" },
{  15,     "Output Vertical Tab Disposition",                "RFC657,JBP" },
{  16,     "Output Linefeed Disposition",                    "RFC657,JBP" },
{  17,     "Extended ASCII",                                 "RFC698,JBP" },
{  18,     "Logout",                                         "RFC727,MRC" },
{  19,     "Byte Macro",                                     "RFC735,JBP" },
{  20,     "Data Entry Terminal",                    "RFC1043,RFC732,JBP" },
{  21,     "Unknown",                                              "None" },
{  22,     "SUPDUP",                                  "RFC736,RFC734,MRC" },
{  22,     "SUPDUP Output",                                  "RFC749,MRC" },
{  23,     "Send Location",                                 "RFC779,EAK1" },
{  24,     "Terminal Type",                                "RFC1091,MS56" },
{  25,     "End of Record",                                  "RFC885,JBP" },
{  26,     "TACACS User Identification",                     "RFC927,BA4" },
{  27,     "Output Marking",                                 "RFC933,SXS" },
{  28,     "Terminal Location Number",                       "RFC946,RN6" },
{  29,     "Telnet 3270 Regime",                            "RFC1041,JXR" },
{  30,     "X.3 PAD",                                      "RFC1053,SL70" },
{  31,     "Negotiate About Window Size",                 "RFC1073,DW183" },
{  32,     "Terminal Speed",                               "RFC1079,CLH3" },
{  33,     "Remote Flow Control",                          "RFC1372,CLH3" },
{  34,     "Linemode",                                     "RFC1184,DB14" },
{  35,     "X Display Location",                           "RFC1096,GM23" },
{  36,     "Environment Option",                           "RFC1408,DB14" },
{  37,     "Authentication Option",                        "RFC1409,DB14" },
{  38,     "Encryption Option",                                    "DB14" },
{  39,     "New Environment Option",                       "RFC1572,DB14" },
{  40,     "TN3270E",                                           "RFC1647" }

};

#define MAX_OPTIONS (sizeof(TelnetOptions) / sizeof(telnet_option))

// Session States
typedef enum  {
	TELNET_ERROR = -1,
	TELNET_NO_CONNECTION = 0,
	TELNET_IDLE,
	TELNET_COMMAND_SENT,
	TELNET_RECEIVING,
	TELNET_RECEIVE_COMPLETE,
	
	TELNET_MAX_STATE
} telnet_state;


// Internal data struct used for Telnet sessions
typedef struct  {
	
	unsigned int		TCPhandle;
	telnet_state		State;
	char *				buffer;
	char *				endptr;
	long				bufferSize;
	long				bufferLength;
	int					UsingGA;
	char *				prompt;
	TelnetCallback *	callback;
	
} telnet_session;


struct session_list {
	telnet_session *			ptr;
	struct session_list *		next;
};

static struct session_list *SessionList = NULL;

// Forward reference
static void telnetFreeSessionStruct(telnet_session *ptr);


// longjmp environment buffer
jmp_buf mark;

// Private function used to catch bad pointer access
static void telnetBadPointer(int sig)
{
	// Return to setjmp location
	longjmp(mark, -1);
}

// Private function used to check pointers
static int telnetCheckPointer(void *ptr, unsigned long size)
{
	int num;
	
	if(ptr == NULL) return TELNET_NULL_POINTER;

	signal (SIGSEGV, telnetBadPointer);

	num = setjmp(mark);
	if(num==0)
	{

#ifdef _CVI_DEBUG
		// Try to access pointer to see if it will cause an exception
		// telnetBadPointer() will be called if SIGSEGV is generated
		num = *((int *) ptr);
#endif		
		// Remove signal handler
		signal (SIGSEGV, SIG_DFL);
		
		// Make sure we have read access to the pointer
		if(IsBadReadPtr(ptr, size) != 0) return TELNET_UNINITIALIZED_POINTER;
		
		return TELNET_SUCCESS;
	}
		
	return TELNET_UNINITIALIZED_POINTER;
}

// Make sure a session pointer is valid
static int telnetCheckSessionPointer(telnet_session *ptr)
{
	int status, found = 0;
	struct session_list *listPtr = SessionList;
	
	status = telnetCheckPointer(ptr, sizeof(telnet_session));
	if(status != TELNET_SUCCESS) return status;
	
	// Make sure point is one telnet.fp created
	while(listPtr != NULL)
	{
		if(listPtr->ptr == ptr)
		{
			found = 1;
			break;
		}
		
		listPtr = listPtr->next;
	}
		
	if(found) return TELNET_SUCCESS;
	
	// Pointer was not found in session pointer list
	return TELNET_UNINITIALIZED_POINTER;
}

// Close all open telnet sessions
static void CleanUpOpenTelnetSessions(void)
{
	struct session_list *listPtr = SessionList;
	
	while(listPtr != NULL)
	{
		if(listPtr->ptr->State != TELNET_NO_CONNECTION)
			DisconnectFromTCPServer (listPtr->ptr->TCPhandle);

		telnetFreeSessionStruct(listPtr->ptr);

		listPtr = listPtr->next;
	}
}

// Create new session struct
static int telnetNewSessionPtr(telnet_session **Session)
{
	static int AtExitInstalled = 0;
	telnet_session *NewSession;
	struct session_list *newListItem, *listPtr;
	
	// Install atexit handler if not already done
	if(!AtExitInstalled)
	{
		atexit(CleanUpOpenTelnetSessions);
		AtExitInstalled++;
	}

	// Allocate telnet session struct memory
	*Session = NULL;
	NewSession = calloc (1, sizeof(telnet_session));
	if(NewSession == NULL) return TELNET_MEMORY_ALLOC_ERROR;
	
	// Set state value
	NewSession->State = TELNET_NO_CONNECTION;
	
	// Allocate memory for buffer	
	NewSession->buffer = calloc(START_BUFFER_SIZE, sizeof(char));
	if(NewSession->buffer == NULL)
	{
		telnetFreeSessionStruct(NewSession);
		return TELNET_MEMORY_ALLOC_ERROR;
	}
	
	// Initilize struct variables
	NewSession->endptr = NewSession->buffer;
	NewSession->bufferSize = START_BUFFER_SIZE;
	NewSession->bufferLength = 0;


	// Allocate new list item
	newListItem = calloc (1, sizeof(struct session_list));
	if(newListItem == NULL)
	{
		telnetFreeSessionStruct(NewSession);
		return TELNET_MEMORY_ALLOC_ERROR;
	}

	// Save pointer
	newListItem->ptr = NewSession;

	// Find end oflist
	listPtr = SessionList;
	while(listPtr != NULL)
	{
		if(listPtr->next == NULL) break;
		listPtr = listPtr->next;
	}


	// Add new item to list
	if(listPtr == NULL)
		SessionList = newListItem;
	else
		// Set up link
		listPtr->next = newListItem;
	
	*Session = NewSession;
	
	return TELNET_SUCCESS;
}


// Private function to clear buffer
static void telnetClearBuffer(telnet_session *Session)
{
	Session->endptr = Session->buffer;
	memset (Session->buffer, 0, Session->bufferSize);
	Session->bufferLength = 0;
}

// Private function to respond to new commands set by server
static void telnetParseCommands(telnet_session *Session, unsigned char *start, int length)
{
	unsigned char command[64], *inPtr, *outPtr;
	int commandLen;
	
	// Scan for options and command data
	inPtr = start;
	while(length > 0)
	{
		// Set output pointer to start of array
		outPtr = command;

		if(*inPtr == IAC)
		{
			*outPtr = *inPtr;
			commandLen = 1;
			
			// Move pointer forward to command
			if(--length == 0) return;
			*(++outPtr) = *(++inPtr);
			commandLen++;

			switch(*inPtr)
			{
				// See if command is a subnegotioation
				case SB:
					// Copy command until subneg end is found
					while(*inPtr != SE)
					{
						if(--length == 0) return;
						*(++outPtr) = *(++inPtr);
						commandLen++;
					}
					*(++outPtr) = SE;
					commandLen++;

					// Null terminate
					*(++outPtr) = 0;
					break;
					
				case GO_AHEAD:
					// Received Go Ahead signal
					if(*(outPtr - 1) == IAC)
						Session->State = TELNET_RECEIVE_COMPLETE;
					*(++outPtr) = 0;
					break;
			
				default:
					// Move pointer forward to Option value
					if(--length == 0) return;
					*(++outPtr) = *(++inPtr);
					commandLen++;
					
					// Null terminate
					*(++outPtr) = 0;
					break;
			}
			
			if(Session->callback)
			{
				// Send command to callback for processing
				Session->callback((TelnetSession)Session, TELNET_CMD_RECEIVE, command);
			}
			else
			{
				// Accept Go Ahead marking, reject all others
				if(commandLen==3)
				{
					if(command[2] == GO_AHEAD)
					{
						switch (command[1])
						{
							case DO:
								telnet_SendCommand ((TelnetSession) Session, WONT, OPT_SUPPRESS_GO_AHEAD);
								telnet_SendCommand ((TelnetSession) Session, DONT, OPT_SUPPRESS_GO_AHEAD);
								// Set flag that we are using GA marking
								Session->UsingGA = 1;
								break;

							case WILL:
								telnet_SendCommand ((TelnetSession) Session, DONT, OPT_SUPPRESS_GO_AHEAD);
								// Set flag that we are using GA marking
								Session->UsingGA = 1;
								break;
						}
					}
					else
					{
						// Reject all other commands found
						switch (command[1])
						{
							case DONT:
							case DO:
								telnet_SendCommand ((TelnetSession) Session, WONT, command[2]);
								break;

							case WILL:
								telnet_SendCommand ((TelnetSession) Session, DONT, command[2]);
								break;
						}
					}
				}
			}
				
			// Erase this command from oringial buffer
			memset(inPtr - (commandLen - 1), 0, commandLen);
		}

		length--;
		inPtr++;
	}
}

// Private function to search for prompt in buffer
static int telnetGotPrompt(telnet_session *Session)
{
	if(Session->prompt == NULL) return 0;

	if(FindPattern (Session->buffer, 0, Session->bufferLength,
					Session->prompt, 1, 1) < 0)
		return 0;
		
	Session->State = TELNET_RECEIVE_COMPLETE;
	return 1;
}


// Private function to read waiting Telnet data
static int telnetRead(telnet_session *Session, unsigned int timeout)
{
	char tempbuffer[MIN_BUFFER_SIZE];
	int i, inBytes, outBytes, status;
	long availableBytes;
	char *inChar, *outChar, *newblock;
	
	// Set curent state value
	Session->State = TELNET_RECEIVING;
	
	// Read in available data
	inBytes = 0;
	do
	{
		// Read in any waiting TCP data
		status = ClientTCPRead (Session->TCPhandle, &tempbuffer[inBytes],
								MIN_BUFFER_SIZE - inBytes, timeout);
		
		// Check for errors. timeout is okay
		if(status < 0 && status != -kTCP_TimeOutErr)
		{
			Session->State = TELNET_ERROR;
			return status;
		}

		inBytes += (status < 0) ? 0: status;
		timeout = MORE_DATA_TIMEOUT;

	// look for more incoming data
	}while(status != -kTCP_TimeOutErr);
		
	if(status == -kTCP_TimeOutErr && inBytes==0) return status;
	
	// Look for commands in new data
	telnetParseCommands(Session, (unsigned char *)tempbuffer, inBytes);
		
	// Check remaining buffer size
	// Increase size if necessary
	availableBytes = Session->bufferSize - Session->bufferLength;
	if(availableBytes < inBytes + 1)
	{
		// Increase by twice the minimum buffer size
		availableBytes = Session->bufferSize + inBytes + MIN_BUFFER_SIZE;
		newblock = realloc (Session->buffer, availableBytes);
	
		// Check for erros
		if(newblock == NULL)
			return TELNET_MEMORY_ALLOC_ERROR;
		
		// Adjust parameters for new memory block size
		Session->bufferSize = availableBytes;
		Session->buffer = newblock;
		Session->endptr = &newblock[Session->bufferLength];
		
		// Clear new portion of memory block
		availableBytes -= Session->bufferLength;
		memset (Session->endptr, 0, availableBytes);
	}
		
	// Transfer non-command data to the session buffer
	// Command data is snipped out by setting it to 0
	for(i=0, outBytes = 0, inChar = tempbuffer, outChar = Session->endptr;
		i<inBytes; 
		i++, inChar++)
	{
		if(*inChar != 0)
		{
			*outChar = *inChar;
			outChar++;
			outBytes++;
		}
	}
		
	// Update pointer and counter in session struct
	Session->endptr = outChar;
	Session->bufferLength += outBytes;


							
	// See if we got a prompt
	if(!Session->UsingGA)
		if(telnetGotPrompt(Session))
			Session->State = TELNET_RECEIVE_COMPLETE;
	
	return TELNET_SUCCESS;
}



// Private function used to handle all TCP callbacks
static int telnetCallback(unsigned int handle, int xType, int errCode, void *callbackData)
{
	telnet_session *Session = (telnet_session *) callbackData;
	char *NewData;
	
	if(telnetCheckSessionPointer(Session) != TELNET_SUCCESS) return -1;
	
	NewData = Session->endptr;
	
	switch(xType)
	{
		case TCP_DISCONNECT:
			DisconnectFromTCPServer (Session->TCPhandle);
			Session->State = TELNET_NO_CONNECTION;

			if(Session->callback != NULL)
				Session->callback((TelnetSession)Session, TELNET_DISCONNECT, NULL);

			break;
			
		case TCP_DATAREADY:
			telnetRead(Session, RECEIVE_TIMEOUT);
			break;
	}
	
		
	return 0;
}



// Private function used to free telnet_session struct memory
static void telnetFreeSessionStruct(telnet_session *ptr)
{
	int found = 0;
	struct session_list *thisItem, *prevItem;
	
	if(ptr == NULL) return;

	// First, shut off the callback
	ptr->callback = NULL;
	
	// Free buffer memory
	if(ptr->buffer != NULL)
		free(ptr->buffer);
		
	// free prompt memory
	if(ptr->prompt != NULL)
		free(ptr->prompt);
		
	// Search for session in list
	thisItem = SessionList;
	prevItem = NULL;
	while(thisItem != NULL)
	{
		if(thisItem->ptr == ptr)
		{
			found = 1;
			break;
		}
		prevItem = thisItem;
		thisItem = thisItem->next;
	}
	
	// Trim the session out of the list
	if(found)
	{
		// Fix up link
		if(prevItem != NULL)
			prevItem->next = thisItem->next;
			
		// If this is the first item,
		// set start of list to second item
		if(thisItem == SessionList)
				SessionList = thisItem->next;

		// Free this list item
		free(thisItem);
	}

	// Free the struct itself
	free(ptr);
}





int __stdcall telnet_OpenSession 
(
	TelnetSession *sessionHandle, 
	char server[],
	TelnetCallback *callbackPointer	
)
{
	int status;
	telnet_session *NewSession;
	
	// Do some error checking
	//
	// Check session handle
	if(sessionHandle == NULL) return TELNET_NULL_POINTER;
	status = telnetCheckPointer(sessionHandle, sizeof(char));
	if(status != TELNET_SUCCESS) return status;
	*sessionHandle = NULL;

	// Check server pointer
	status = telnetCheckPointer(server, sizeof(char));
	if(status != TELNET_SUCCESS) return status;
	
	// Allocate memory for session info
	//
	// Allocate telnet_session structure
	status = telnetNewSessionPtr(&NewSession);
	if(status != TELNET_SUCCESS) return status;

	// Set callback pointer
	telnet_RegisterCallback(NewSession, callbackPointer);
	
	
	// Open a connection to the server
	status = ConnectToTCPServer (&NewSession->TCPhandle, TELNET_PORT, server,
								 telnetCallback, NewSession,
								 CONNECT_TIMEOUT);
	if(status != 0)
	{
		telnetFreeSessionStruct(NewSession);
		return status;
	}

	// Set disconnect mode
	status = SetTCPDisconnectMode (NewSession->TCPhandle,
								   TCP_DISCONNECT_MANUAL);
	if(status != 0)
	{
		telnetFreeSessionStruct(NewSession);
		return status;
	}
	
	
	NewSession->State = TELNET_IDLE;
	
	// Fire the callback if needed
	if(NewSession->callback != NULL)
		NewSession->callback((TelnetSession)NewSession, TELNET_CONNECT, NULL);
	else
		telnet_SendCommand ((TelnetSession)NewSession, DONT, OPT_SUPPRESS_GO_AHEAD);

	*sessionHandle = NewSession;
	
	return TELNET_SUCCESS;
}

// Function to perform telnet log in
int __stdcall telnet_LogIn (TelnetSession sessionHandle, 
							char LoginPrompt[], char PasswordPrompt[],
							char userName[], char password[])
{
	int status, bytes;
	double startTime, timeout;
	char *endptr;
	unsigned int handle;
	telnet_session *Session;
	
	// Do some error checking
	//
	// Check session handle
	if(sessionHandle == NULL) return TELNET_NULL_POINTER;
	status = telnetCheckSessionPointer(sessionHandle);
	if(status != TELNET_SUCCESS) return status;

	// Check login prompt pointer
	status = telnetCheckPointer(LoginPrompt, sizeof(char));
	if(status != TELNET_SUCCESS) return status;
	
	// Check password prompt pointer
	status = telnetCheckPointer(PasswordPrompt, sizeof(char));
	if(status != TELNET_SUCCESS) return status;

	// Check user name pointer
	status = telnetCheckPointer(userName, sizeof(char));
	if(status != TELNET_SUCCESS) return status;
	
	// Check password pointer
	status = telnetCheckPointer(password, sizeof(char));
	if(status != TELNET_SUCCESS) return status;

	// Make correct datatype copy of session handle
	// just to make code more readable
	Session = (telnet_session *) sessionHandle;
	
	// Set up some parameters
	handle = Session->TCPhandle;
	timeout = PROMPT_TIMEOUT / 1000;

	// Look for login prompt
	startTime = Timer();
	while(strstr(Session->buffer, LoginPrompt) == NULL)
	{
		// Check for timeout
		if(Timer() - startTime > timeout)
		{
			Session->State = TELNET_ERROR;
			return TELNET_PROMPT_TIMEOUT;
		}
		
		// Save current buffer position
		endptr = Session->endptr;
		bytes = Session->bufferLength;
		
		// Allow system to process any incoming
		// network traffic
		ProcessSystemEvents();
		
		// Check to see if we got any new data
		// If we got some data, parse any telnet commands in it
		bytes = Session->bufferLength - bytes;
		telnetParseCommands(Session, (unsigned char *) endptr, bytes);
	}
		
	// Send the user name
	status = ClientTCPWrite (handle, userName, strlen(userName), PROMPT_TIMEOUT);
	if(status < 0)
	{
		Session->State = TELNET_ERROR;
		return status;
	}

	
	// Wait for password prompt
	startTime = Timer();
	ProcessSystemEvents();
	while(strstr(Session->buffer, PasswordPrompt) == NULL)
	{
		if(Timer() - startTime > timeout)
		{
			Session->State = TELNET_ERROR;
			return TELNET_PROMPT_TIMEOUT;
		}
		ProcessSystemEvents();
	}
	
	// Send the password
	status = ClientTCPWrite (handle, password, strlen(password), PROMPT_TIMEOUT);
	if(status < 0)
	{
		Session->State = TELNET_ERROR;
		return status;
	}
	
	// Wait until we stop receiving data
	startTime = Timer();
	while(Timer() - startTime < timeout)
	{
		// Save current buffer position
		endptr = Session->endptr;
		
		// Allow system to process any incoming
		// network traffic
		ProcessSystemEvents();
		
		// Check to see if we got any new data
		if(endptr != Session->endptr)
		{
			// Reset timeout and start time to shorter values
			timeout = (MORE_DATA_TIMEOUT << 1) / 1000.0;
			startTime = Timer();
		}
	}
	
	// Check for bad login
	if(Session->State == TELNET_NO_CONNECTION
		|| Session->State == TELNET_ERROR
		|| strstr(Session->buffer, "ncorrect") != NULL)
	{
		return TELNET_BAD_LOGIN;
	}
	
	// Not using Go-Ahead option, get prompt
	return telnet_GetPrompt(sessionHandle);
}


int __stdcall telnet_CloseSession (TelnetSession *sessionHandle)
{
	int status;
	telnet_session *Session; 
	
	status = telnetCheckPointer(sessionHandle, sizeof(char));
	if(status != TELNET_SUCCESS) return status;

	Session = *((telnet_session **) sessionHandle);
	
	status = telnetCheckSessionPointer(Session);
	if(status != TELNET_SUCCESS) return status;

	status = DisconnectFromTCPServer (Session->TCPhandle);
	
	if(Session->callback != NULL)
		Session->callback((TelnetSession)Session, TELNET_DISCONNECT, NULL);
				
	telnetFreeSessionStruct(Session);
	*sessionHandle = NULL;
	
	return status;
}

// Function to find telnet prompt
int __stdcall telnet_GetPrompt(TelnetSession sessionHandle)
{
	int status;
	char *start, *ptr;
	telnet_session *Session;
	
	// Check pointer
	status = telnetCheckSessionPointer(sessionHandle);
	if(status != TELNET_SUCCESS) return status;

	Session = (telnet_session *) sessionHandle;
	
	// Clear out incoming data
	do
	{
		ptr = Session->endptr;
		status = telnetRead(Session, MORE_DATA_TIMEOUT);
		if(status < 0 && status != -kTCP_TimeOutErr)
		{
			Session->State = TELNET_ERROR;
			return status;
		}
	
	}while(ptr != Session->endptr);


	// Hit return
	status = ClientTCPWrite (Session->TCPhandle, "\r", 1, MORE_DATA_TIMEOUT);
	if(status < 0)
	{
		Session->State = TELNET_ERROR;
		return status;
	}

	// Clear the session receive buffer
	telnetClearBuffer(Session);
	
	// Read the response
	ptr = Session->endptr;
	status = telnetRead(Session, MORE_DATA_TIMEOUT);
	if(status < 0)
	{
		Session->State = TELNET_ERROR;
		return status;
	}
	

	// Get rid of paths, shell ids, and escape chars
	for(status=0, start=Session->buffer;
		status < strlen(start);
		status++, start++)
	{
		if(*start==ESC)
		{
			*start = ' ';
			while(*(++start) != 'H')
				*start = ' ';
		}
	}
	
	start = Session->buffer;
	while(*start == ' ') start++;
	
	ptr = strrchr (start, '{');
	if(ptr != NULL)
		*ptr = 0;

	ptr = strchr (start, '\\');
	if(ptr != NULL)
		*ptr = 0;

	ptr = strchr (start, '\\');
	if(ptr != NULL)
		*ptr = 0;

	
	if(Session->prompt != NULL)
	{
		if(strcmp (Session->prompt, start) == 0)
			return TELNET_SUCCESS;

		free(Session->prompt);
		Session->prompt = NULL;
	}
			
			
	status = strlen(start);
	if(status < 1)
	{
		Session->State = TELNET_ERROR;
		return status;
	}

	Session->prompt = calloc (status + 1, sizeof(char));
	if(Session->prompt == NULL)
		return TELNET_MEMORY_ALLOC_ERROR;
			
	
	memcpy (Session->prompt, start, status);
	
	Session->State = TELNET_IDLE;
	
	return TELNET_SUCCESS;
}


int __stdcall telnet_SendText (TelnetSession sessionHandle, char command[])
{
	static const unsigned char cmdGA[2] = { IAC, GO_AHEAD };
	int status;
	telnet_session *Session;

	// Check pointer
	status = telnetCheckSessionPointer(sessionHandle);
	if(status != TELNET_SUCCESS) return status;

	Session = (telnet_session *) sessionHandle;
	
	// TODO: Add state check and interrupt if
	// another command is still executing
	
	// Clear the current buffer
	telnetClearBuffer(Session);
	
	// Send the text
	status = ClientTCPWrite (Session->TCPhandle, command, strlen(command), PROMPT_TIMEOUT);
	if(status < 0)
	{
		Session->State = TELNET_ERROR;
		return status;
	}
	
	if(Session->UsingGA)
	{
		// Send the Go-Ahead command
		status = ClientTCPWrite (Session->TCPhandle, cmdGA, 2, PROMPT_TIMEOUT);
		if(status < 0)
		{
			Session->State = TELNET_ERROR;
			return status;
		}
	}
	
	Session->State = TELNET_COMMAND_SENT;
	
	ProcessSystemEvents();
	
	return TELNET_SUCCESS;
}

int __stdcall telnet_WaitForResponse (	TelnetSession sessionHandle,
                            		 	char *responseString[],
                            		 	double timeout)
{
	int status;
	double startTime;
	unsigned int Len;
	telnet_session *Session;

	// Check pointer
	status = telnetCheckSessionPointer(sessionHandle);
	if(status != TELNET_SUCCESS) return status;

	Session = (telnet_session *) sessionHandle;
	
	startTime = Timer();
	while(Timer() - startTime < timeout)
	{
		if(Session->State == TELNET_RECEIVE_COMPLETE) break;
		ProcessSystemEvents();
	}
	
	if(Session->State != TELNET_RECEIVE_COMPLETE)
		return TELNET_PROMPT_TIMEOUT;
		
	// Copy current buffer
	Len = Session->bufferLength;
	if(Len == 0)
		*responseString = NULL;
	else
	{
		*responseString = calloc (Len + 1, sizeof(char));
		memcpy (*responseString, Session->buffer, Len);
	}

	return TELNET_SUCCESS;
}

int __stdcall telnet_SendCommand (TelnetSession sessionHandle,
                                 unsigned char action, unsigned char option)
{
	int status;
	unsigned char command[3];
	telnet_session *Session;
	
	// Check pointer
	status = telnetCheckSessionPointer(sessionHandle);
	if(status != TELNET_SUCCESS) return status;

	Session = (telnet_session *) sessionHandle;
	
	if(action != DO && action != DONT && action != WILL && action != WONT)
		return TELNET_BAD_PARAMETER;
		
	if(option > MAX_OPTIONS)
		return TELNET_BAD_PARAMETER;
		
	// Build command
	command[0] = IAC;
	command[1] = action;
	command[2] = option;
	
	ClientTCPWrite (Session->TCPhandle, command, 3, PROMPT_TIMEOUT);
		
	if(option == OPT_SUPPRESS_GO_AHEAD)
	{
		if(action == DONT)
			Session->UsingGA = 1;
		else if(action == DO)
			Session->UsingGA = 0;
	}
		
	// Fire the callback if needed
	if(Session->callback != NULL)
		Session->callback((TelnetSession)Session, TELNET_CMD_RESPONSE, command);
	
	return TELNET_SUCCESS;
}

int __stdcall telnet_SendOption (TelnetSession sessionHandle,
                                unsigned char option, unsigned char *optionData,
                                int optionDataLength)
{
	int status;
	unsigned char *OptionCommand;
	telnet_session *Session;
	
	// Check pointer
	status = telnetCheckSessionPointer(sessionHandle);
	if(status != TELNET_SUCCESS) return status;

	Session = (telnet_session *) sessionHandle;
	
	if(option > MAX_OPTIONS)
		return TELNET_BAD_PARAMETER;
		
	if(optionDataLength < 1 || optionDataLength > 255)
		return TELNET_BAD_PARAMETER;
		
	OptionCommand = calloc(optionDataLength + 5, sizeof(unsigned char));
	if(OptionCommand == NULL)
		return TELNET_MEMORY_ALLOC_ERROR;
		
	// Build a option command
	OptionCommand[0] = IAC;
	OptionCommand[1] = SB;
	OptionCommand[2] = option;
	memcpy(&OptionCommand[3], optionData, optionDataLength);
	OptionCommand[3 + optionDataLength] = IAC;
	OptionCommand[4 + optionDataLength] = SE;
	
	ClientTCPWrite (Session->TCPhandle, OptionCommand, optionDataLength + 5, PROMPT_TIMEOUT);
		
	// Fire the callback if needed
	if(Session->callback != NULL)
		Session->callback((TelnetSession)Session, TELNET_CMD_RESPONSE, OptionCommand);
	
	return TELNET_SUCCESS;
}


int __stdcall telnet_RegisterCallback (TelnetSession sessionHandle,
                                      TelnetCallback callbackPointer)
{
	int status;
	telnet_session *Session;

	// Check session pointer
	status = telnetCheckSessionPointer(sessionHandle);
	if(status != TELNET_SUCCESS) return status;

	Session = (telnet_session *) sessionHandle;
	
	// Check callback pointer, NULL is okay here
	status = telnetCheckPointer(callbackPointer, sizeof(char));
	if(status != TELNET_SUCCESS && 
		status != TELNET_NULL_POINTER) return status;

	// Save callback pointer
	Session->callback = callbackPointer;
	
	return TELNET_SUCCESS;
}

const char * __stdcall telnet_GetErrorText (int errorCode)
{

	switch(errorCode)
	{
		case TELNET_SUCCESS:
			return "Success";

		case TELNET_NULL_POINTER:
			return "Received null pointer";
			
		case TELNET_UNINITIALIZED_POINTER:
			return "Received uninitialized pointer";
			
		case TELNET_BAD_PARAMETER:
			return "Bad parameter passed to function";
			
		case TELNET_MEMORY_ALLOC_ERROR:
			return "Memory allocation error";
			
		case TELNET_PROMPT_TIMEOUT:
			return "Timeout waiting for prompt";
			
		case TELNET_BAD_LOGIN:
			return "Login incorrect";
			
		case TELNET_UNKNOWN_OPTION:
			return "Unknown Telnet Option";
			
		case TELNET_TCP_READ_ERROR:
			return "TCP/IP read error";
			
		default:
			// Return TCP error message if any
			if(errorCode > -1000)
				return GetTCPErrorString (errorCode);
			else
				return "Unknown error";
	}
}


int __stdcall telnet_GetOptionInfo (unsigned char option, const char *description[],
                                    const char *RFC[])
{
	static int NumOptions = MAX_OPTIONS - 1;
	
    if(option > NumOptions)
    {
    	if(telnetCheckPointer(description, sizeof(char)) == TELNET_SUCCESS)
    		*description = NULL;
    	
    	if(telnetCheckPointer(RFC, sizeof(char)) == TELNET_SUCCESS)
    		*RFC = NULL;
    	return TELNET_UNKNOWN_OPTION;
    }
    
	if(telnetCheckPointer(description, sizeof(char)) == TELNET_SUCCESS)
		*description = TelnetOptions[option].description;
		
	if(telnetCheckPointer(RFC, sizeof(char)) == TELNET_SUCCESS)	
		*RFC = TelnetOptions[option].RFC;
	
	return TELNET_SUCCESS;
}


// DLL entry point
//
#ifdef _CVI_DLL_

int __stdcall DllMain (HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
	switch (fdwReason)
	{
		case DLL_PROCESS_ATTACH:
			if (InitCVIRTE (hinstDLL, 0, 0) == 0)
				return 0;	  /* out of memory */
			break;
		case DLL_PROCESS_DETACH:
			CloseCVIRTE ();
			break;
	}
	
	return 1;
}

#endif

