/*======================================================================
* Original Author: Matt Kropp, June 2000
*
* All files and code in this projected are
* subject to the GNU General Public License.
* A copy of this license may be found here:
* http://www.gnu.org/licenses/gpl.html
*
* Notes:
* Although Telnet looks simple from a 
* user's point of view, it is an extremely
* complex protocol. This code does not
* attempt to be a complete telnet client.
* Instead, it attempts to to the bare minimum
* required to control a telnet session from
* within an application.
*
========================================================================
Module:		Telnet Header File

Description:
Contains prototypes and declarations for INI FIle Function Panel
========================================================================*/

#include <cvidef.h>

#ifndef _TELNET_FP_
#define _TELNET_FP_


////////////////////////////////////////////////////////////////////////////
// IETF defined Telnet Codes
#define BELL					7
#define BACKSPACE				8
#define HORIZ_TAB				9
#define VERT_TAB				11
#define FORM_FEED				12
#define ESC						27		/* Escape char */

#define NUL						0		/* Same as NOP */
#define IS						0		/* Used with Status option RFC859 */
#define SEND					1		/* Used with Status option RFC859 */


// Telnet Options
#define OPT_BINARY				0		/* Transmit Binary RFC856 */
#define OPT_ECHO				1		/* Echo chars RFC857 */
#define OPT_SUPPRESS_GO_AHEAD	3		/* Suppress use of Go-Ahead marker RFC858 */
#define OPT_STATUS				5		/* Request status of current options RFC859 */
#define OPT_TIMING_MARK			6		/* Request a timing mark to be sent RFC860 */
#define OPT_REMOTE_CONTROL		7		/* Remote Controlled Trans and Echo RFC726 */
#define OPT_NAOCRD				10		/* Negotiate About Output Carriage Return Disposition RFC652 */
#define OPT_HTS					11		/* Output Horizontal Tab Stops RFC653 */
#define OPT_NAOHTD				12		/* Negotiate About Output Horizontal Tab Disposition RFC654 */
#define OPT_NAOFFD				13		/* Negotiate About Output Formfeed Disposition RFC655 */
#define OPT_VT					14		/* Output Vertical Tab Stops RFC656 */
#define OPT_NAOVTD				15		/* Negotiate About Output Vertical Tab Stops RFC657 */
#define OPT_NAOLFD				16		/* Negotiate About Output Linefeed Disposition RFC658 */
#define OPT_EXTEND_ASCII		17		/* Allow extended ascii chars RFC698 */
#define OPT_LOGOUT				18		/* Logout options RFC727 */
#define OPT_BM					19		/* Define Byte Macro RFC735 */
#define OPT_DET					20		/* Data Entry Terminal option RFC732 */
#define OPT_SEND_LOCATION		23		/* Send Location option RFC779 */
#define OPT_TERMINAL_TYPE		24		/* Set terminal type RFC930 */
#define OPT_END_OF_RECORD		25		/* End of record marking (IAC EOR) RFC885 */
#define OPT_TUID				26		/* TACACS User ID RFC927 */
#define OPT_OUTPUT_MARK			27		/* Display banner on each page of data RFC933 */
#define OPT_TERMINAL_SPEED		32		/* Terminal Speed option RFC1079 */
#define OPT_LINEMODE			34		/* Line mode RFC1184 */
#define OPT_AUTHENTICATION		37		/* Authentication option RFC1409 */
#define OPT_ENCRYPTION			38		/* Encryption option */

// Telnet data
#define EOR						239		/* End of record marker */
#define	SE						240		/* End subnegotiation parameters */
#define NOP						241
#define DATA_MARK				242
#define BREAK					243
#define INTERRUPT				244
#define ABORT_OUTPUT			245
#define ARE_YOU_THERE			246
#define ERASE_CHAR				247
#define ERASE_LINE				248
#define GO_AHEAD				249
#define SB						250		/* Start of subnegotiation of option parameters */
#define WILL					251		/* Will do option */
#define WONT					252		/* Won't do option */
#define DO						253		/* Do requested option */
#define DONT					254		/* Don't or stop doing option */
#define IAC						255		/* Interprete As Command, start/end of command marker*/


////////////////////////////////////////////////////////////////////////////
// Event Codes
typedef enum {
	TELNET_DISCONNECT = 1,
	TELNET_CONNECT,
	TELNET_RECEIVE,
	TELNET_SEND,
	TELNET_CMD_RECEIVE,
	TELNET_CMD_RESPONSE
} telnet_event;


////////////////////////////////////////////////////////////////////////////
// Error codes
#define TELNET_SUCCESS					0
#define TELNET_NULL_POINTER				-1001
#define TELNET_UNINITIALIZED_POINTER	-1002
#define TELNET_BAD_PARAMETER			-1003
#define TELNET_MEMORY_ALLOC_ERROR		-1004
#define TELNET_PROMPT_TIMEOUT			-1005
#define TELNET_BAD_LOGIN				-1006
#define TELNET_UNKNOWN_OPTION			-1007
#define TELNET_TCP_READ_ERROR			-1008


////////////////////////////////////////////////////////////////////////////
// Defines used by telnet_SendCommandResponse
#define TELNET_WILL_DO					1
#define TELNET_WONT_DONT				0

////////////////////////////////////////////////////////////////////////////
// Type definition for telnet session handle
typedef void *	TelnetSession;

////////////////////////////////////////////////////////////////////////////
// Callback function type definition
typedef void (TelnetCallback)(TelnetSession sessionHandle, telnet_event TelnetEvent, const unsigned char EventData[]);
								
////////////////////////////////////////////////////////////////////////////								
// Function definitions
int __stdcall telnet_OpenSession (	TelnetSession *sessionHandle, char server[],
									TelnetCallback callbackPointer);
                                 
int __stdcall telnet_LogIn (TelnetSession sessionHandle, 
							char LoginPrompt[], char PasswordPrompt[],
							char userName[], char password[]);
							
int __stdcall telnet_CloseSession (TelnetSession *sessionHandle);

int __stdcall telnet_GetPrompt(TelnetSession sessionHandle);

int __stdcall telnet_SendCommand (TelnetSession sessionHandle,
                                  unsigned char action, unsigned char option);

int __stdcall telnet_SendOption (TelnetSession sessionHandle,
                                 unsigned char option, unsigned char *optionData,
                                 int optionDataLength);

int __stdcall telnet_SendText (TelnetSession sessionHandle, char text[]);

int __stdcall telnet_WaitForResponse (	TelnetSession sessionHandle,
										char *responseString[], double timeOut);
                            
int __stdcall telnet_RegisterCallback (	TelnetSession sessionHandle,
										TelnetCallback callbackPointer);
                             
const char *  __stdcall telnet_GetErrorText (int errorCode);

int __stdcall telnet_GetOptionInfo (unsigned char option, 
									const char *description[],
									const char *RFC[]);

#endif	//  _TELNET_FP_
